/*  Sjaak, a program for playing chess variants
 *  Copyright (C) 2011  Evert Glebbeek
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef bitboard_t_H
#define bitboard_t_H

#include <stdint.h>
#include "assert.h"
#include "square.h"
#include "squares.h"
#include "bits.h"

typedef uint64_t bitboard_t;
#ifdef __i386__
typedef uint128_t __attribute__((aligned(8))) large_bitboard_t;
#else
typedef uint128_t large_bitboard_t;
#endif
typedef union {bitboard_t bb; large_bitboard_t lbb;} ubitboard_t;

/* Special bitboards, used for masking and move generation:
 * The "board_empty" and "board_all" bitboards speak for themselves.
 * The "board_minus_(a|h)" boards are used in move generation when
 * shifting pawn bit boards.
 * The centre, xcentre, kwing and qwing boards can be used in tactical
 * situations to specifically look for moves in that part of the board.
 * Rank encodings are mainly used in move generation, but could also be
 * used in function evaluations (think rooks on the 7th row).
 * The "castlek" and "castleq" bitboards are bitmasks that have to fit the
 * king, rook and inibit bitboards; used in move generation. Their "i"
 * counter parts mark the squares that have to be empty.
 */
extern bitboard_t board_empty;
extern bitboard_t board_all;
extern bitboard_t board_edge;
extern bitboard_t board_minus_a;
extern bitboard_t board_minus_h;
extern bitboard_t board_dark;
extern bitboard_t board_light;
extern bitboard_t board_rank1;
extern bitboard_t board_rank2;
extern bitboard_t board_rank3;
extern bitboard_t board_rank4;
extern bitboard_t board_rank5;
extern bitboard_t board_rank6;
extern bitboard_t board_rank7;
extern bitboard_t board_rank8;
extern bitboard_t board_afile;
extern bitboard_t board_bfile;
extern bitboard_t board_cfile;
extern bitboard_t board_dfile;
extern bitboard_t board_efile;
extern bitboard_t board_ffile;
extern bitboard_t board_gfile;
extern bitboard_t board_hfile;

extern bitboard_t short_castle_mask[2];
extern bitboard_t short_castle_free[2];
extern bitboard_t short_castle_safe[2];
extern bitboard_t long_castle_mask[2];
extern bitboard_t long_castle_free[2];
extern bitboard_t long_castle_safe[2];

extern int8_t short_castle_king_dest[2];
extern int8_t long_castle_king_dest[2];

/* Array-based indexing for rank and file bitboards */
static const bitboard_t board_rank[8] = {
   0x00000000000000FFll, 0x000000000000FF00ll,
   0x0000000000FF0000ll, 0x00000000FF000000ll,
   0x000000FF00000000ll, 0x0000FF0000000000ll,
   0x00FF000000000000ll, 0xFF00000000000000ll,
};

static const bitboard_t board_file[8] = {
   0x0101010101010101ll, 0x0202020202020202ll,
   0x0404040404040404ll, 0x0808080808080808ll,
   0x1010101010101010ll, 0x2020202020202020ll,
   0x4040404040404040ll, 0x8080808080808080ll,
};

/* Diagonal masks */
static const bitboard_t board_a1h8_diagonal[16] = {
   H1_MASK,
   G1_MASK|H2_MASK,
   F1_MASK|G2_MASK|H3_MASK,
   E1_MASK|F2_MASK|G3_MASK|H4_MASK,
   D1_MASK|E2_MASK|F3_MASK|G4_MASK|H5_MASK,
   C1_MASK|D2_MASK|E3_MASK|F4_MASK|G5_MASK|H6_MASK,
   B1_MASK|C2_MASK|D3_MASK|E4_MASK|F5_MASK|G6_MASK|H7_MASK,
   A1_MASK|B2_MASK|C3_MASK|D4_MASK|E5_MASK|F6_MASK|G7_MASK|H8_MASK,
   A2_MASK|B3_MASK|C4_MASK|D5_MASK|E6_MASK|F7_MASK|G8_MASK,
   A3_MASK|B4_MASK|C5_MASK|D6_MASK|E7_MASK|F8_MASK,
   A4_MASK|B5_MASK|C6_MASK|D7_MASK|E8_MASK,
   A5_MASK|B6_MASK|C7_MASK|D8_MASK,
   A6_MASK|B7_MASK|C8_MASK,
   A7_MASK|B8_MASK,
   A8_MASK,
};

static const bitboard_t board_a8h1_diagonal[16] = {
   A1_MASK,
   B1_MASK|A2_MASK,
   C1_MASK|B2_MASK|A3_MASK,
   D1_MASK|C2_MASK|B3_MASK|A4_MASK,
   E1_MASK|D2_MASK|C3_MASK|B4_MASK|A5_MASK,
   F1_MASK|E2_MASK|D3_MASK|C4_MASK|B5_MASK|A6_MASK,
   G1_MASK|F2_MASK|E3_MASK|D4_MASK|C5_MASK|B6_MASK|A7_MASK,
   H1_MASK|G2_MASK|F3_MASK|E4_MASK|D5_MASK|C6_MASK|B7_MASK|A8_MASK,
   H2_MASK|G3_MASK|F4_MASK|E5_MASK|D6_MASK|C7_MASK|B8_MASK,
   H3_MASK|G4_MASK|F5_MASK|E6_MASK|D7_MASK|C8_MASK,
   H4_MASK|G5_MASK|F6_MASK|E7_MASK|D8_MASK,
   H5_MASK|G6_MASK|F7_MASK|E8_MASK,
   H6_MASK|G7_MASK|F8_MASK,
   H7_MASK|G8_MASK,
   H8_MASK,
};

extern const bitboard_t square_bitboards[64];

/* Large bitboards */
extern large_bitboard_t large_square_bitboards[128];
extern large_bitboard_t large_board_empty;
extern large_bitboard_t large_board_all;
extern large_bitboard_t large_board_edge;
extern large_bitboard_t large_board_east_edge;
extern large_bitboard_t large_board_west_edge;
extern large_bitboard_t large_board_north_edge;
extern large_bitboard_t large_board_south_edge;
extern large_bitboard_t large_board_light;
extern large_bitboard_t large_board_dark;
extern large_bitboard_t large_board_rank[16];
extern large_bitboard_t large_board_file[16];
extern large_bitboard_t large_board_south;
extern large_bitboard_t large_board_north;
extern large_bitboard_t large_board_northward[16];
extern large_bitboard_t large_board_southward[16];

extern large_bitboard_t large_board_diagonal[32];
extern large_bitboard_t large_board_antidiagonal[32];

extern large_bitboard_t large_short_castle_mask[2];
extern large_bitboard_t large_short_castle_free[2];
extern large_bitboard_t large_short_castle_safe[2];
extern large_bitboard_t large_long_castle_mask[2];
extern large_bitboard_t large_long_castle_free[2];
extern large_bitboard_t large_long_castle_safe[2];

extern uint16_t large_row_mask;

extern void initialise_bitboards(int files, int ranks);

extern void printf_bitboard(bitboard_t b);
extern void print_bitboard_squares(bitboard_t b);

extern void initialise_large_bitboards(int files, int ranks);
extern void printf_large_bitboard(large_bitboard_t b);
extern void print_large_bitboard_squares(large_bitboard_t b);

/*******************************************/
/* Bitboard manipulation routines (inline) */
/*******************************************/

/* Set a bit on the bitboard */
static inline void set_bitboard(bitboard_t *board, const int square)
{
   assert(board);
   (*board) |= ((uint64_t)1)<<square;
}

/* ANDs two bitboards */
static inline bitboard_t and_bitboards(const bitboard_t b1, const bitboard_t b2)
{
   return b1&b2;
}

/* MASKs two bitboards: set bits in first bitboard to 0 if set in second */
static inline bitboard_t mask_bitboards(const bitboard_t b1, const bitboard_t b2)
{
   return b1&~b2;
}

/* ORs two bitboards */
static inline bitboard_t or_bitboards(const bitboard_t b1, const bitboard_t b2)
{
   return b1|b2;
}

/* Retrieve one row from a bitboard */
static inline uint8_t get_bitboard_row(const bitboard_t b, const int row)
{
   return b >> (row*8);
}

/* Return a bitboard representation of a given square */
static inline bitboard_t make_bitboard_square(const int square)
{
#if defined(__LP64__)
   return ((uint64_t)1)<<square;
#else
   return square_bitboards[square];
#endif
}

/* Expand one row into a full bitboard */
static inline bitboard_t make_bitboard_row(const uint8_t data, const int row_nr)
{
   return ((uint64_t)data) << (row_nr*8);
}

/* Expand one packed file into a full bitboard */
static inline bitboard_t make_bitboard_file(const uint8_t data, const int file_nr)
{
   bitboard_t b;
   int n;
   b = 0;
   for(n=0;n<8;n++) {
      if (data & (1<<n)) {
         set_bitboard(&b, pack_row_file(n, file_nr));
      }
   }
   return b;
}

/* Shift a bitboard to the right (that is, a->b, b->c etc.) by 1 file */
/*  and discards what gets shifted in on the left */
/* Note that shifting the board to the right implies shifting the binary */
/*  values to the left, since files to the right have a higher index */
static inline bitboard_t shift_bitboard_file_right(const bitboard_t b)
{
   return and_bitboards(board_minus_a, (b<<1));
}

/* Shift a bitboard to the left (that is, b->a, c->b etc.) by 1 file */
/*  and discards what gets shifted in on the right */
/* Note that shifting the board to the left implies shifting the binary */
/*  values to the right, since files to the left have a lower index */
static inline bitboard_t shift_bitboard_file_left(const bitboard_t b)
{
   return and_bitboards(board_minus_h, (b>>1));
}

/* Shift a bitboard up one row (that is, 1->2, 2->3 etc */
static inline bitboard_t shift_bitboard_row_up(const bitboard_t b)
{
   return b<<8;
}

/* Shift a bitboard down one row (that is, 1->2, 2->3 etc */
static inline bitboard_t shift_bitboard_row_down(const bitboard_t b)
{
   return b>>8;
}

/* Combinations of the above */
static inline bitboard_t shift_bitboard_right_up(const bitboard_t b)
{
   return shift_bitboard_row_up(shift_bitboard_file_right(b));
}

static inline bitboard_t shift_bitboard_right_down(const bitboard_t b)
{
   return shift_bitboard_row_down(shift_bitboard_file_right(b));
}

static inline bitboard_t shift_bitboard_left_up(const bitboard_t b)
{
   return shift_bitboard_row_up(shift_bitboard_file_left(b));
}

static inline bitboard_t shift_bitboard_left_down(const bitboard_t b)
{
   return shift_bitboard_row_down(shift_bitboard_file_left(b));
}


/*************************************************/
/* Large bitboard manipulation routines (inline) */
/*************************************************/

/* Retrieve one row from a bitboard */
static inline uint16_t get_large_bitboard_row(large_bitboard_t b, const int row)
{
   typedef union { uint128_t i128; uint64_t i64[2]; } uuint128_t;
   b = shr128(b, row * large_board_files);
   return ((uuint128_t)b).i64[0] & large_row_mask;
}

/* Retrieve one file from a bitboard */
static inline uint16_t get_large_bitboard_file(large_bitboard_t b, const int file)
{
   uint16_t file_bits = 0;
   int n;

   int bit = file;
   for (n=0; n<large_board_ranks; n++) {
      if (test128(b, bit)) file_bits |= (1<<n);
      bit += large_board_files;
   }
   return file_bits;
}

static inline large_bitboard_t shift_large_bitboard_file_right(const large_bitboard_t b)
{
   return shl128((~large_board_east_edge) & b, 1);
}

static inline large_bitboard_t shift_large_bitboard_file_left(const large_bitboard_t b)
{
   return shr128((~large_board_west_edge) & b, 1);
}

#endif
